import pylibmc

from remotecv.utils import logger
from remotecv.result_store import BaseStore

class ResultStore(BaseStore):

    WEEK = 604800
    memcache_instance = None

    def __init__(self, config):
        if not ResultStore.memcache_instance:
            host_strings = config.memcache_hosts.split(',')
            ResultStore.memcache_instance = pylibmc.Client(
                host_strings,
                binary=True,
                behaviors={
                    "tcp_nodelay": True,
                    'no_block': True,
                    "ketama": True
                }
            )
        self.storage = ResultStore.memcache_instance

    def store(self, key, points):
        result = self.serialize(points)
        logger.debug("Points found: %s" % result)
        key = "thumbor-detector-%s" % key
        self.storage.set(key, result, time=2*self.WEEK)



import cv

class Image:
    @classmethod
    def create_from_raw_bytes(cls, size, mode, img_data):
        instance = cls()
        instance.set_image_data(size, mode, img_data)
        return instance

    @classmethod
    def create_from_buffer(cls, image_buffer):
        instance = cls()
        if not instance.is_valid(image_buffer):
            return None

        instance.set_image_buffer(image_buffer)
        return instance

    def is_valid(self, image_buffer):
        return len(image_buffer) > 4 and image_buffer[:4] != 'GIF8'

    def set_image_buffer(self, image_buffer):
        buffer_len = len(image_buffer)
        imagefiledata = cv.CreateMatHeader(1, buffer_len, cv.CV_8UC1)
        cv.SetData(imagefiledata, image_buffer, buffer_len)
        self.image = cv.DecodeImage(imagefiledata, cv.CV_LOAD_IMAGE_COLOR)
        self.size = cv.GetSize(self.image)
        self.mode = "BGR"

    def set_image_raw_data(self, size, mode, img_data):
        self.image = cv.CreateImageHeader(size, cv.IPL_DEPTH_8U, 3)
        cv.SetData(self.image, img_data)
        self.mode = mode
        self.size = size

    def grayscale(self):
        convert_mode = getattr(cv, 'CV_%s2GRAY' % self.mode)

        gray = cv.CreateImage(self.size, cv.IPL_DEPTH_8U, 1)
        cv.CvtColor(self.image, gray, convert_mode)
        cv.EqualizeHist(gray, gray)
        self.image = gray



#!/usr/bin/python
# -*- coding: utf-8 -*-

# remote cv service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

import sys
import argparse
import logging

from redis import Redis

from remotecv.utils import config

def import_module(name):
    module = __import__(name)
    if '.' in name:
        module = reduce(getattr, name.split('.')[1:], module)
    return module

def start_pyres_worker():
    from remotecv.unique_queue import UniqueWorker
    redis = Redis(host=config.redis_host, port=config.redis_port, password=config.redis_password)
    UniqueWorker.run(['Detect'], redis, timeout=config.timeout)

def start_celery_worker():
    from remotecv.celery_tasks import CeleryTasks

    celery_tasks = CeleryTasks(config.key_id, config.key_secret, config.region, config.timeout, config.polling_interval)
    celery_tasks.run_commands(config.extra_args, log_level=config.log_level)

def main(params=None):
    if params is None:
        params = sys.argv[1:]
    parser = argparse.ArgumentParser(description='Runs RemoteCV.')

    conn_group = parser.add_argument_group('Worker Backend')
    conn_group.add_argument('-b', '--backend', default='pyres', choices=['pyres', 'celery'], help='Worker backend')

    conn_group = parser.add_argument_group('Pyres Connection Arguments')
    conn_group.add_argument('--host', default='localhost', help='Redis host')
    conn_group.add_argument('--port', default=6379, type=int, help='Redis port')
    conn_group.add_argument('--password', default=None, help='Redis password')

    conn_group = parser.add_argument_group('Celery/SQS Connection Arguments')
    conn_group.add_argument('--region', default='us-east-1', help='AWS SQS Region')
    conn_group.add_argument('--key_id', default='', help='AWS access key id')
    conn_group.add_argument('--key_secret', default='', help='AWS access key secret')
    conn_group.add_argument('--polling_interval', default=20, help='AWS polling interval')

    other_group = parser.add_argument_group('Other arguments')
    other_group.add_argument('-l', '--level', default='debug', help='Logging level')
    other_group.add_argument('-o', '--loader', default='remotecv.http_loader', help='Loader used')
    other_group.add_argument('-s', '--store', default='remotecv.result_store.redis_store', help='Loader used')
    other_group.add_argument('-t', '--timeout', default=None, type=int, help='Timeout in seconds for image detection')

    memcache_store_group = parser.add_argument_group('Memcache store arguments')
    memcache_store_group.add_argument('--memcache_hosts', default='localhost:11211', help='Comma separated list of memcache hosts')

    parser.add_argument('args', nargs=argparse.REMAINDER)

    arguments = parser.parse_args(params)
    logging.basicConfig(level=getattr(logging, arguments.level.upper()))

    config.backend = arguments.backend
    config.redis_host = arguments.host
    config.redis_port = arguments.port
    config.redis_password = arguments.password

    config.region = arguments.region
    config.key_id = arguments.key_id
    config.key_secret = arguments.key_secret
    config.polling_interval = arguments.polling_interval

    config.timeout = arguments.timeout
    config.log_level = arguments.level.upper()
    config.loader = import_module(arguments.loader)
    config.store = import_module(arguments.store)

    config.memcache_hosts = arguments.memcache_hosts

    config.extra_args = sys.argv[:1] + arguments.args

    if arguments.backend == 'pyres':
        start_pyres_worker()
    elif arguments.backend == 'celery':
        start_celery_worker()

if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license: 
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

from remotecv.detectors import CascadeLoaderDetector

HAIR_OFFSET = 0.12

class FaceDetector(CascadeLoaderDetector):

    def __init__(self):
        self.load_cascade_file(__file__, 'haarcascade_frontalface_alt.xml')

    def __add_hair_offset(self, top, height):
        top = max(0, top - height * HAIR_OFFSET)
        return top

    def detect(self, image):
        features = self.get_features(image)

        points = []
        if features:
            for (left, top, width, height), neighbors in features:
                top = self.__add_hair_offset(top, height)
                points.append([left, top, width, height])

        return points



import json

class BaseStore(object):

    def __init__(self, config):
        self.config = config

    def serialize(self, points):
        result_array = []
        for point in points:
            result = {
                'x': point[0] + (int(point[2]) / 2),
                'y': point[1] + (int(point[3]) / 2),
                'height': int(point[2]),
                'width': int(point[3]),
                'origin': ''
            }
            result['z'] = result['width'] * result['height']
            result_array.append(result)

        return json.dumps(result_array)

    def store(self, key, points):
        raise NotImplementedError()



from celery import Celery
from pyres_tasks import DetectTask


class CeleryTasks:

    def __init__(self, key_id, key_secret, region, timeout=None, polling_interval=None):
        self.celery = Celery(broker='sqs://%s:%s@' % (key_id, key_secret))

        self.celery.conf.update(
            BROKER_TRANSPORT_OPTIONS = {
                'region': region,
                'visibility_timeout': timeout or 120,
                'polling_interval': polling_interval or 20,
                'queue_name_prefix': 'celery-remotecv-'
            }
        )

    def get_detect_task(self):
        @self.celery.task(ignore_result=True, acks_late=True)
        def detect_task(detection_type, image_path, key):
            DetectTask.perform(detection_type, image_path, key)

        return detect_task

    def run_commands(self, args, log_level=None):
        # We have to init the task so it can be found by the worker later
        self.get_detect_task()

        if log_level:
            self.celery.conf.update(CELERYD_LOG_LEVEL=log_level)
        self.celery.start(args)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license: 
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

from remotecv.detectors import CascadeLoaderDetector


class ProfileDetector(CascadeLoaderDetector):

    def __init__(self):
        self.load_cascade_file(__file__, 'haarcascade_profileface.xml')



from redis import Redis

from remotecv.utils import logger
from remotecv.result_store import BaseStore

class ResultStore(BaseStore):

    WEEK = 604800
    redis_instance = None

    def __init__(self, config):
        if not ResultStore.redis_instance:
            ResultStore.redis_instance = Redis(host=config.redis_host, port=config.redis_port, password=config.redis_password)
        self.storage = ResultStore.redis_instance

    def store(self, key, points):
        result = self.serialize(points)
        logger.debug("Points found: %s" % result)
        redis_key = "thumbor-detector-%s" % key
        self.storage.setex(redis_key, result, 2 * self.WEEK)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license: 
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

from setuptools import setup, find_packages
from remotecv import __version__

setup(
    name = 'remotecv',
    version = __version__,
    description = "remotecv is an OpenCV worker for facial and feature recognition",
    long_description = """
remotecv is an OpenCV worker for facial and feature recognition
""",
    keywords = 'imaging face detection feature opencv',
    author = 'globo.com',
    author_email = 'timehome@corp.globo.com',
    url = 'https://github.com/globocom/remotecv/wiki',
    license = 'MIT',
    classifiers = ['Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Natural Language :: English',
                   'Operating System :: MacOS',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 2.6',
                   'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
                   'Topic :: Multimedia :: Graphics :: Presentation'
    ],
    packages = find_packages(),
    package_data = {
        '': ['*.xml'],
    },

    install_requires=[
        "argparse>=1.2.1,<1.3.0",
        "pyres>=1.2,<1.3"
    ],

    entry_points = {
        'console_scripts': [
            'remotecv = remotecv.worker:main',
            'remotecv-web = remotecv.web:main'
        ],
    }
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

from os.path import join, dirname, abspath, isabs

import cv

class BaseDetector(object):

    def detect(self, context):
        raise NotImplementedError()


class CascadeLoaderDetector(BaseDetector):

    def load_cascade_file(self, module_path, cascade_file_path):
        if not hasattr(self.__class__, 'cascade'):
            if isabs(cascade_file_path):
                cascade_file = cascade_file_path
            else:
                cascade_file = join(abspath(dirname(module_path)), cascade_file_path)
            setattr(self.__class__, 'cascade', cv.Load(cascade_file))

    def get_min_size_for(self, size):
        ratio = int(min(size[0], size[1]) / 15)
        ratio = max(20, ratio)
        return (ratio, ratio)

    def get_features(self, image):
        min_size = self.get_min_size_for(image.size)
        haar_scale = 1.2
        min_neighbors = 4

        faces = cv.HaarDetectObjects(image.image,
                                     self.__class__.cascade, cv.CreateMemStorage(0),
                                     haar_scale, min_neighbors,
                                     cv.CV_HAAR_DO_CANNY_PRUNING, min_size)

        faces_scaled = []

        for ((x, y, w, h), n) in faces:
            # the input to cv.HaarDetectObjects was resized, so scale the
            # bounding box of each face and convert it to two CvPoints
            pt1 = (x, y)
            pt2 = ((x + w), (y + h))
            x1 = pt1[0]
            x2 = pt2[0]
            y1 = pt1[1]
            y2 = pt2[1]
            faces_scaled.append(((x1, y1, x2-x1, y2-y1), None))

        return faces_scaled

    def detect(self, image):
        features = self.get_features(image)

        if features:
            points = [[left, top, width, height] for (left, top, width, height), neighbors in features]
        else:
            points = []

        return points



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/globocom/thumbor/wiki

# Licensed under the MIT license: 
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com timehome@corp.globo.com

import cv

from remotecv.detectors import BaseDetector

class FeatureDetector(BaseDetector):

    def detect(self, image):
        rows, cols = image.size

        eig_image = cv.CreateMat(rows, cols, cv.CV_32FC1)
        temp_image = cv.CreateMat(rows, cols, cv.CV_32FC1)
        points = cv.GoodFeaturesToTrack(image.image, eig_image, temp_image, 20, 0.04, 1.0, useHarris = False)

        if points:
            return [[x, y, 1, 1] for x, y in points]

        return None
